@@ -35,8 +35,11 @@ Context managers for use with the ``with`` statement.
 
 from contextlib import contextmanager, nested
 import sys
+import socket
+import select
 
-from fabric.state import output, win32
+from fabric.thread_handling import ThreadHandler
+from fabric.state import output, win32, connections, env
 from fabric import state
 
 if not win32:
@@ -462,6 +465,105 @@ def shell_env(**kw):
     return _setenv({'shell_env': kw})
 
 
+def _forwarder(chan, sock):
+    # Bidirectionally forward data between a socket and a Paramiko channel.
+    while True:
+        r, w, x = select.select([sock, chan], [], [])
+        if sock in r:
+            data = sock.recv(1024)
+            if len(data) == 0:
+                break
+            chan.send(data)
+        if chan in r:
+            data = chan.recv(1024)
+            if len(data) == 0:
+                break
+            sock.send(data)
+    chan.close()
+    sock.close()
+
+
+@documented_contextmanager
+def remote_tunnel(remote_port, local_port=None, local_host="localhost",
+    remote_bind_address="127.0.0.1"):
+    """
+    Create a tunnel forwarding a locally-visible port to the remote target.
+
+    For example, you can let the remote host access a database that is
+    installed on the client host::
+
+        # Map localhost:6379 on the server to localhost:6379 on the client,
+        # so that the remote 'redis-cli' program ends up speaking to the local
+        # redis-server.
+        with remote_tunnel(6379):
+            run("redis-cli -i")
+
+    The database might be installed on a client only reachable from the client
+    host (as opposed to *on* the client itself)::
+
+        # Map localhost:6379 on the server to redis.internal:6379 on the client
+        with remote_tunnel(6379, local_host="redis.internal")
+            run("redis-cli -i")
+
+    ``remote_tunnel`` accepts up to four arguments:
+
+    * ``remote_port`` (mandatory) is the remote port to listen to.
+    * ``local_port`` (optional) is the local port to connect to; the default is
+      the same port as the remote one.
+    * ``local_host`` (optional) is the locally-reachable computer (DNS name or
+      IP address) to connect to; the default is ``localhost`` (that is, the
+      same computer Fabric is running on).
+    * ``remote_bind_address`` (optional) is the remote IP address to bind to
+      for listening, on the current target. It should be an IP address assigned
+      to an interface on the target (or a DNS name that resolves to such IP).
+      You can use "0.0.0.0" to bind to all interfaces.
+
+    .. note::
+        By default, most SSH servers only allow remote tunnels to listen to the
+        localhost interface (127.0.0.1). In these cases, `remote_bind_address`
+        is ignored by the server, and the tunnel will listen only to 127.0.0.1.
+    """
+    if local_port is None:
+        local_port = remote_port
+
+    sockets = []
+    channels = []
+    threads = []
+
+    def accept(channel, (src_addr, src_port), (dest_addr, dest_port)):
+        channels.append(channel)
+        sock = socket.socket()
+        sockets.append(sock)
+
+        try:
+            sock.connect((local_host, local_port))
+        except Exception, e:
+            print "[%s] rtunnel: cannot connect to %s:%d (from local)" % (env.host_string, local_host, local_port)
+            chan.close()
+            return
+
+        print "[%s] rtunnel: opened reverse tunnel: %r -> %r -> %r"\
+              % (env.host_string, channel.origin_addr,
+                 channel.getpeername(), (local_host, local_port))
+
+        th = ThreadHandler('fwd', _forwarder, channel, sock)
+        threads.append(th)
+
+    transport = connections[env.host_string].get_transport()
+    transport.request_port_forward(remote_bind_address, remote_port, handler=accept)
+
+    try:
+        yield
+    finally:
+        for sock, chan, th in zip(sockets, channels, threads):
+            sock.close()
+            chan.close()
+            th.thread.join()
+            th.raise_if_needed()
+        transport.cancel_port_forward(remote_bind_address, remote_port)
+
+
+
 quiet = lambda: settings(hide('everything'), warn_only=True)
 quiet.__doc__ = """
     Alias to ``settings(hide('everything'), warn_only=True)``.
