@@ -29,6 +29,7 @@ from fabric.utils import (
     indent,
     _pty_size,
     warn,
+    apply_lcwd
 )
 
 
@@ -274,11 +275,6 @@ def put(local_path=None, remote_path=None, use_sudo=False,
         contents of the file-like object by rewinding it using ``seek`` (and
         will use ``tell`` afterwards to preserve the previous file position).
 
-    .. note::
-        Use of a file-like object in `~fabric.operations.put`'s ``local_path``
-        argument will cause a temporary file to be utilized due to limitations
-        in our SSH layer's API.
-
     ``remote_path`` may also be a relative or absolute location, but applied to
     the remote host. Relative paths are relative to the remote user's home
     directory, but tilde expansion (e.g. ``~/.ssh/``) will also be performed if
@@ -361,12 +357,9 @@ def put(local_path=None, remote_path=None, use_sudo=False,
             remote_path = env.cwd.rstrip('/') + '/' + remote_path
 
         if local_is_path:
-            # Expand local paths
+            # Apply lcwd, expand tildes, etc
             local_path = os.path.expanduser(local_path)
-            # Honor lcd() where it makes sense
-            if not os.path.isabs(local_path) and env.lcwd:
-                local_path = os.path.join(env.lcwd, local_path)
-
+            local_path = apply_lcwd(local_path, env)
             # Glob local path
             names = glob(local_path)
         else:
@@ -488,14 +481,6 @@ def get(remote_path, local_path=None):
         not be restored to its previous location, as that doesn't make as much
         sense here and/or may not even be possible.
 
-    .. note::
-        Due to how our SSH layer works, a temporary file will still be written
-        to your hard disk even if you specify a file-like object such as a
-        StringIO for the ``local_path`` argument. Cleanup is performed,
-        however -- we just note this for users expecting straight-to-memory
-        transfers. (We hope to patch our SSH layer in the future to enable true
-        straight-to-memory downloads.)
-
     .. note::
         If a file-like object such as StringIO has a ``name`` attribute, that
         will be used in Fabric's printed output instead of the default
@@ -527,8 +512,8 @@ def get(remote_path, local_path=None):
         and callable(local_path.write))
 
     # Honor lcd() where it makes sense
-    if local_is_path and not os.path.isabs(local_path) and env.lcwd:
-        local_path = os.path.join(env.lcwd, local_path)
+    if local_is_path:
+        local_path = apply_lcwd(local_path, env)
 
     ftp = SFTP(env.host_string)
 
@@ -568,15 +553,12 @@ def get(remote_path, local_path=None):
                     result = ftp.get_dir(remote_path, local_path)
                     local_files.extend(result)
                 else:
-                    # Result here can be file contents (if not local_is_path)
-                    # or final resultant file path (if local_is_path)
+                    # Perform actual get. If getting to real local file path,
+                    # add result (will be true final path value) to
+                    # local_files. File-like objects are omitted.
                     result = ftp.get(remote_path, local_path, local_is_path,
                         os.path.basename(remote_path))
-                    if not local_is_path:
-                        # Overwrite entire contents of local_path
-                        local_path.seek(0)
-                        local_path.write(result)
-                    else:
+                    if local_is_path:
                         local_files.append(result)
 
         except Exception, e:
@@ -679,11 +661,11 @@ def _prefix_env_vars(command, local=False):
     path = env.path
     if path:
         if env.path_behavior == 'append':
-            path = '$PATH:\"%s\" ' % path
+            path = '$PATH:\"%s\"' % path
         elif env.path_behavior == 'prepend':
-            path = '\"%s\":$PATH ' % path
+            path = '\"%s\":$PATH' % path
         elif env.path_behavior == 'replace':
-            path = '\"%s\" ' % path
+            path = '\"%s\"' % path
 
         env_vars['PATH'] = path
 
@@ -709,7 +691,7 @@ def _prefix_env_vars(command, local=False):
 
 
 def _execute(channel, command, pty=True, combine_stderr=None,
-    invoke_shell=False, stdout=None, stderr=None):
+    invoke_shell=False, stdout=None, stderr=None, timeout=None):
     """
     Execute ``command`` over ``channel``.
 
@@ -732,6 +714,12 @@ def _execute(channel, command, pty=True, combine_stderr=None,
     stdout = stdout or sys.stdout
     stderr = stderr or sys.stderr
 
+    # Timeout setting control
+    timeout = env.command_timeout if (timeout is None) else timeout
+
+    # What to do with CTRl-C?
+    remote_interrupt = env.remote_interrupt
+
     with char_buffered(sys.stdin):
         # Combine stdout and stderr to get around oddball mixing issues
         if combine_stderr is None:
@@ -761,7 +749,7 @@ def _execute(channel, command, pty=True, combine_stderr=None,
             if command:
                 channel.sendall(command + "\n")
         else:
-            channel.exec_command(command)
+            channel.exec_command(command=command)
 
         # Init stdout, stderr capturing. Must use lists instead of strings as
         # strings are immutable and we're using these as pass-by-reference
@@ -771,12 +759,17 @@ def _execute(channel, command, pty=True, combine_stderr=None,
 
         workers = (
             ThreadHandler('out', output_loop, channel, "recv",
-                capture=stdout_buf, stream=stdout),
+                capture=stdout_buf, stream=stdout, timeout=timeout),
             ThreadHandler('err', output_loop, channel, "recv_stderr",
-                capture=stderr_buf, stream=stderr),
+                capture=stderr_buf, stream=stderr, timeout=timeout),
             ThreadHandler('in', input_loop, channel, using_pty)
         )
 
+        if remote_interrupt is None:
+            remote_interrupt = invoke_shell
+        if remote_interrupt and not using_pty:
+            remote_interrupt = False
+
         while True:
             if channel.exit_status_ready():
                 break
@@ -786,7 +779,12 @@ def _execute(channel, command, pty=True, combine_stderr=None,
                 # exception within, recv_exit_status())
                 for worker in workers:
                     worker.raise_if_needed()
-            time.sleep(ssh.io_sleep)
+            try:
+                time.sleep(ssh.io_sleep)
+            except KeyboardInterrupt:
+                if not remote_interrupt:
+                    raise
+                channel.send('\x03')
 
         # Obtain exit code of remote program now that we're done.
         status = channel.recv_exit_status()
@@ -860,7 +858,7 @@ def _noop():
 
 def _run_command(command, shell=True, pty=True, combine_stderr=True,
     sudo=False, user=None, quiet=False, warn_only=False, stdout=None,
-    stderr=None, group=None):
+    stderr=None, group=None, timeout=None):
     """
     Underpinnings of `run` and `sudo`. See their docstrings for more info.
     """
@@ -890,7 +888,7 @@ def _run_command(command, shell=True, pty=True, combine_stderr=True,
         result_stdout, result_stderr, status = _execute(
             channel=default_channel(), command=wrapped_command, pty=pty,
             combine_stderr=combine_stderr, invoke_shell=False, stdout=stdout,
-            stderr=stderr)
+            stderr=stderr, timeout=timeout)
 
         # Assemble output string
         out = _AttributeString(result_stdout)
@@ -900,7 +898,7 @@ def _run_command(command, shell=True, pty=True, combine_stderr=True,
         out.failed = False
         out.command = given_command
         out.real_command = wrapped_command
-        if status != 0:
+        if status not in env.ok_ret_codes:
             out.failed = True
             msg = "%s() received nonzero return code %s while executing" % (
                 which, status
@@ -928,7 +926,7 @@ def _run_command(command, shell=True, pty=True, combine_stderr=True,
 
 @needs_host
 def run(command, shell=True, pty=True, combine_stderr=None, quiet=False,
-    warn_only=False, stdout=None, stderr=None):
+    warn_only=False, stdout=None, stderr=None, timeout=None):
     """
     Run a shell command on a remote host.
 
@@ -980,11 +978,17 @@ def run(command, shell=True, pty=True, combine_stderr=None, quiet=False,
     could even provide your own stream objects or loggers, e.g. ``myout =
     StringIO(); run("command", stdout=myout)``.
 
+    If you want an exception raised when the remote program takes too long to
+    run, specify ``timeout=N`` where ``N`` is an integer number of seconds,
+    after which to time out. This will cause ``run`` to raise a
+    `~fabric.exceptions.CommandTimeout` exception.
+
     Examples::
 
         run("ls /var/www/")
         run("ls /home/myuser", shell=False)
         output = run('ls /var/www/site1')
+        run("take_a_long_time", timeout=5)
 
     .. versionadded:: 1.0
         The ``succeeded`` and ``stderr`` return value attributes, the
@@ -1003,14 +1007,18 @@ def run(command, shell=True, pty=True, combine_stderr=None, quiet=False,
 
     .. versionadded:: 1.5
         The return value attributes ``.command`` and ``.real_command``.
+
+    .. versionadded:: 1.6
+        The ``timeout`` argument.
     """
     return _run_command(command, shell, pty, combine_stderr, quiet=quiet,
-        warn_only=warn_only, stdout=stdout, stderr=stderr)
+        warn_only=warn_only, stdout=stdout, stderr=stderr, timeout=timeout)
 
 
 @needs_host
 def sudo(command, shell=True, pty=True, combine_stderr=None, user=None,
-         quiet=False, warn_only=False, stdout=None, stderr=None, group=None):
+    quiet=False, warn_only=False, stdout=None, stderr=None, group=None,
+    timeout=None):
     """
     Run a shell command on a remote host, with superuser privileges.
 
@@ -1053,8 +1061,9 @@ def sudo(command, shell=True, pty=True, combine_stderr=None, user=None,
     return _run_command(
         command, shell, pty, combine_stderr, sudo=True,
         user=user if user else env.sudo_user,
-        group=group, quiet=quiet,
-        warn_only=warn_only, stdout=stdout, stderr=stderr)
+        group=group, quiet=quiet, warn_only=warn_only, stdout=stdout,
+        stderr=stderr, timeout=timeout
+    )
 
 
 def local(command, capture=False, shell=None):
@@ -1139,7 +1148,7 @@ def local(command, capture=False, shell=None):
     out.failed = False
     out.return_code = p.returncode
     out.stderr = err
-    if p.returncode != 0:
+    if p.returncode not in env.ok_ret_codes:
         out.failed = True
         msg = "local() encountered an error (return code %s) while executing '%s'" % (p.returncode, command)
         error(message=msg, stdout=out, stderr=err)
