@@ -11,14 +11,15 @@ import unittest
 import random
 import types
 
-from nose.tools import raises, eq_
+from nose.tools import raises, eq_, ok_
 from fudge import with_patched_object
 
 from fabric.state import env, output
 from fabric.operations import require, prompt, _sudo_prefix, _shell_wrap, \
     _shell_escape
-from fabric.api import get, put, hide, show, cd, lcd, local, run, sudo
+from fabric.api import get, put, hide, show, cd, lcd, local, run, sudo, quiet
 from fabric.sftp import SFTP
+from fabric.exceptions import CommandTimeout
 
 from fabric.decorators import with_settings
 from utils import *
@@ -104,12 +105,70 @@ def test_require_noniterable_provided_by_key():
     require('foo', provided_by=fake_providing_function)
 
 
+@aborts
+def test_require_key_exists_empty_list():
+    """
+    When given a single existing key but the value is an empty list, require()
+    aborts
+    """
+    # 'hosts' is one of the default values, so we know it'll be there
+    require('hosts')
+
+
+@aborts
+@with_settings(foo={})
+def test_require_key_exists_empty_dict():
+    """
+    When given a single existing key but the value is an empty dict, require()
+    aborts
+    """
+    require('foo')
+
+
+@aborts
+@with_settings(foo=())
+def test_require_key_exists_empty_tuple():
+    """
+    When given a single existing key but the value is an empty tuple, require()
+    aborts
+    """
+    require('foo')
+
+
+@aborts
+@with_settings(foo=set())
+def test_require_key_exists_empty_set():
+    """
+    When given a single existing key but the value is an empty set, require()
+    aborts
+    """
+    require('foo')
+
+
+@with_settings(foo=0, bar=False)
+def test_require_key_exists_false_primitive_values():
+    """
+    When given keys that exist with primitive values that evaluate to False,
+    require() throws no exception
+    """
+    require('foo', 'bar')
+
+
+@with_settings(foo=['foo'], bar={'bar': 'bar'}, baz=('baz',), qux=set('qux'))
+def test_require_complex_non_empty_values():
+    """
+    When given keys that exist with non-primitive values that are not empty,
+    require() throws no exception
+    """
+    require('foo', 'bar', 'baz', 'qux')
+
+
 #
 # prompt()
 #
 
 def p(x):
-    print x,
+    sys.stdout.write(x)
 
 
 @mock_streams('stdout')
@@ -144,7 +203,7 @@ def test_sudo_prefix_with_user():
     _sudo_prefix() returns prefix plus -u flag for nonempty user
     """
     eq_(
-        _sudo_prefix(user="foo"),
+        _sudo_prefix(user="foo", group=None),
         "%s -u \"foo\" " % (env.sudo_prefix % env)
     )
 
@@ -153,7 +212,27 @@ def test_sudo_prefix_without_user():
     """
     _sudo_prefix() returns standard prefix when user is empty
     """
-    eq_(_sudo_prefix(user=None), env.sudo_prefix % env)
+    eq_(_sudo_prefix(user=None, group=None), env.sudo_prefix % env)
+
+
+def test_sudo_prefix_with_group():
+    """
+    _sudo_prefix() returns prefix plus -g flag for nonempty group
+    """
+    eq_(
+        _sudo_prefix(user=None, group="foo"),
+        "%s -g \"foo\" " % (env.sudo_prefix % env)
+    )
+
+
+def test_sudo_prefix_with_user_and_group():
+    """
+    _sudo_prefix() returns prefix plus -u and -g for nonempty user and group
+    """
+    eq_(
+        _sudo_prefix(user="foo", group="bar"),
+        "%s -u \"foo\" -g \"bar\" " % (env.sudo_prefix % env)
+    )
 
 
 @with_settings(use_shell=True)
@@ -265,6 +344,69 @@ class TestCombineStderr(FabricTest):
         eq_("stdout", r.stdout)
         eq_("stderr", r.stderr)
 
+
+class TestQuietAndWarnKwargs(FabricTest):
+    @server(responses={'wat': ["", "", 1]})
+    def test_quiet_implies_warn_only(self):
+        # Would raise an exception if warn_only was False
+        eq_(run("wat", quiet=True).failed, True)
+
+    @server()
+    @mock_streams('both')
+    def test_quiet_implies_hide_everything(self):
+        run("ls /", quiet=True)
+        eq_(sys.stdout.getvalue(), "")
+        eq_(sys.stderr.getvalue(), "")
+
+    @server(responses={'hrm': ["", "", 1]})
+    @mock_streams('both')
+    def test_warn_only_is_same_as_settings_warn_only(self):
+        eq_(run("hrm", warn_only=True).failed, True)
+
+    @server()
+    @mock_streams('both')
+    def test_warn_only_does_not_imply_hide_everything(self):
+        run("ls /simple", warn_only=True)
+        assert sys.stdout.getvalue() != ""
+
+
+class TestMultipleOKReturnCodes(FabricTest):
+    @server(responses={'no srsly its ok': ['', '', 1]})
+    def test_expand_to_include_1(self):
+        with settings(quiet(), ok_ret_codes=[0, 1]):
+            eq_(run("no srsly its ok").succeeded, True)
+
+
+slow_server = server(responses={'slow': ['', '', 0, 3]})
+slow = lambda x: slow_server(raises(CommandTimeout)(x))
+
+class TestRun(FabricTest):
+    """
+    @server-using generic run()/sudo() tests
+    """
+    @slow
+    def test_command_timeout_via_env_var(self):
+        env.command_timeout = 2 # timeout after 2 seconds
+        with hide('everything'):
+            run("slow")
+
+    @slow
+    def test_command_timeout_via_kwarg(self):
+        with hide('everything'):
+            run("slow", timeout=2)
+
+    @slow
+    def test_command_timeout_via_env_var_in_sudo(self):
+        env.command_timeout = 2 # timeout after 2 seconds
+        with hide('everything'):
+            sudo("slow")
+
+    @slow
+    def test_command_timeout_via_kwarg_of_sudo(self):
+        with hide('everything'):
+            sudo("slow", timeout=2)
+
+
 #
 # get() and put()
 #
@@ -759,6 +901,22 @@ class TestFileTransfers(FabricTest):
             get(f, f)
         assert self.exists_locally(os.path.join(d, f))
 
+    @server()
+    @mock_streams('stdout')
+    def test_stringio_without_name(self):
+        file_obj = StringIO(u'test data')
+        put(file_obj, '/')
+        assert re.search('<file obj>', sys.stdout.getvalue())
+
+    @server()
+    @mock_streams('stdout')
+    def test_stringio_with_name(self):
+        """If a file object (StringIO) has a name attribute, use that in output"""
+        file_obj = StringIO(u'test data')
+        file_obj.name = 'Test StringIO Object'
+        put(file_obj, '/')
+        assert re.search(file_obj.name, sys.stdout.getvalue())
+
 
 #
 # local()
@@ -788,3 +946,27 @@ def test_local_output_and_capture():
                     local.description = d
                     yield local, "echo 'foo' >/dev/null", capture
                     del local.description
+
+
+class TestRunSudoReturnValues(FabricTest):
+    @server()
+    def test_returns_command_given(self):
+        """
+        run("foo").command == foo
+        """
+        with hide('everything'):
+            eq_(run("ls /").command, "ls /")
+
+    @server()
+    def test_returns_fully_wrapped_command(self):
+        """
+        run("foo").real_command involves env.shell + etc
+        """
+        # FabTest turns use_shell off, we must reactivate it.
+        # Doing so will cause a failure: server's default command list assumes
+        # it's off, we're not testing actual wrapping here so we don't really
+        # care. Just warn_only it.
+        with settings(hide('everything'), warn_only=True, use_shell=True):
+            # Slightly flexible test, we're not testing the actual construction
+            # here, just that this attribute exists.
+            ok_(env.shell in run("ls /").real_command)
